@@ -129,8 +129,8 @@ static inline const struct cpumask *uvm_cpumask_of_node(int node)
     #endif
 
 // See bug 1707453 for further details about setting the minimum kernel version.
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 32)
-#  error This driver does not support kernels older than 2.6.32!
+#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 10, 0)
+#  error This driver does not support kernels older than 3.10!
 #endif
 
 #if !defined(VM_RESERVED)
@@ -146,6 +146,10 @@ static inline const struct cpumask *uvm_cpumask_of_node(int node)
 #define VM_MIXEDMAP    0x00000000
 #endif
 
+#if !defined(MPOL_PREFERRED_MANY)
+#define MPOL_PREFERRED_MANY    5
+#endif
+
 //
 // printk.h already defined pr_fmt, so we have to redefine it so the pr_*
 // routines pick up our version
@@ -210,10 +214,6 @@ static inline const struct cpumask *uvm_cpumask_of_node(int node)
 
 #define NV_UVM_GFP_FLAGS (GFP_KERNEL)
 
-#if !defined(NV_ADDRESS_SPACE_INIT_ONCE_PRESENT)
-    void address_space_init_once(struct address_space *mapping);
-#endif
-
 // Develop builds define DEBUG but enable optimization
 #if defined(DEBUG) && !defined(NVIDIA_UVM_DEVELOP)
   // Wrappers for functions not building correctly without optimizations on,
@@ -345,23 +345,6 @@ static inline NvU64 NV_GETTIME(void)
              (bit) = find_next_zero_bit((addr), (size), (bit) + 1))
 #endif
 
-// bitmap_clear was added in 2.6.33 via commit c1a2a962a2ad103846e7950b4591471fabecece7
-#if !defined(NV_BITMAP_CLEAR_PRESENT)
-    static inline void bitmap_clear(unsigned long *map, unsigned int start, int len)
-    {
-        unsigned int index = start;
-        for_each_set_bit_from(index, map, start + len)
-            __clear_bit(index, map);
-    }
-
-    static inline void bitmap_set(unsigned long *map, unsigned int start, int len)
-    {
-        unsigned int index = start;
-        for_each_clear_bit_from(index, map, start + len)
-            __set_bit(index, map);
-    }
-#endif
-
 // Added in 2.6.24
 #ifndef ACCESS_ONCE
   #define ACCESS_ONCE(x) (*(volatile typeof(x) *)&(x))
@@ -418,6 +401,7 @@ static inline NvU64 NV_GETTIME(void)
 // 654672d4ba1a6001c365833be895f9477c4d5eab ("locking/atomics:
 // Add _{acquire|release|relaxed}() variants of some atomic operations") in v4.3
 // (2015-08-06).
+// TODO: Bug 3849079: We always use this definition on newer kernels.
 #ifndef atomic_read_acquire
     #define atomic_read_acquire(p) smp_load_acquire(&(p)->counter)
 #endif
@@ -426,23 +410,30 @@ static inline NvU64 NV_GETTIME(void)
     #define atomic_set_release(p, v) smp_store_release(&(p)->counter, v)
 #endif
 
+// atomic_long_read_acquire and atomic_long_set_release were added in commit
+// b5d47ef9ea5c5fe31d7eabeb79f697629bd9e2cb ("locking/atomics: Switch to
+// generated atomic-long") in v5.1 (2019-05-05).
+// TODO: Bug 3849079: We always use these definitions on newer kernels.
+#define atomic_long_read_acquire uvm_atomic_long_read_acquire
+static inline long uvm_atomic_long_read_acquire(atomic_long_t *p)
+{
+    long val = atomic_long_read(p);
+    smp_mb();
+    return val;
+}
+
+#define atomic_long_set_release uvm_atomic_long_set_release
+static inline void uvm_atomic_long_set_release(atomic_long_t *p, long v)
+{
+    smp_mb();
+    atomic_long_set(p, v);
+}
 
 // Added in 3.11
 #ifndef PAGE_ALIGNED
     #define PAGE_ALIGNED(addr) (((addr) & (PAGE_SIZE - 1)) == 0)
 #endif
 
-// Added in 2.6.37 via commit e1ca7788dec6773b1a2bce51b7141948f2b8bccf
-#if !defined(NV_VZALLOC_PRESENT)
-    static inline void *vzalloc(unsigned long size)
-    {
-        void *p = vmalloc(size);
-        if (p)
-            memset(p, 0, size);
-        return p;
-    }
-#endif
-
 // Changed in 3.17 via commit 743162013d40ca612b4cb53d3a200dff2d9ab26e
 #if (NV_WAIT_ON_BIT_LOCK_ARGUMENT_COUNT == 3)
     #define UVM_WAIT_ON_BIT_LOCK(word, bit, mode) \
@@ -498,21 +489,6 @@ static bool radix_tree_empty(struct radix_tree_root *tree)
 #endif
 #endif
 
-#if !defined(NV_USLEEP_RANGE_PRESENT)
-static void __sched usleep_range(unsigned long min, unsigned long max)
-{
-    unsigned min_msec = min / 1000;
-    unsigned max_msec = max / 1000;
-
-    if (min_msec != 0)
-        msleep(min_msec);
-    else if (max_msec != 0)
-        msleep(max_msec);
-    else
-        msleep(1);
-}
-#endif
-
 typedef struct
 {
     struct mem_cgroup *new_memcg;
@@ -533,6 +509,7 @@ typedef struct
     }
 
 #if defined(NVCPU_X86) || defined(NVCPU_X86_64)
+  #include <asm/pgtable.h>
   #include <asm/pgtable_types.h>
 #endif
 
@@ -540,6 +517,31 @@ typedef struct
   #define PAGE_KERNEL_NOENC PAGE_KERNEL
 #endif
 
+// uvm_pgprot_decrypted is a GPL-aware version of pgprot_decrypted that returns
+// the given input when UVM cannot use GPL symbols, or pgprot_decrypted is not
+// defined. Otherwise, the function is equivalent to pgprot_decrypted. UVM only
+// depends on pgprot_decrypted when the driver is allowed to use GPL symbols:
+// both AMD's SEV and Intel's TDX are only supported in conjunction with OpenRM.
+//
+// It is safe to invoke uvm_pgprot_decrypted in KVM + AMD SEV-SNP guests, even
+// if the call is not required, because pgprot_decrypted(PAGE_KERNEL_NOENC) ==
+// PAGE_KERNEL_NOENC.
+//
+// pgprot_decrypted was added by commit 21729f81ce8a ("x86/mm: Provide general
+// kernel support for memory encryption") in v4.14 (2017-07-18)
+static inline pgprot_t uvm_pgprot_decrypted(pgprot_t prot)
+{
+#if defined(pgprot_decrypted)
+        // TODO: Bug 3972645: add an assertion to warn that the call to
+        // pgprot_decrypted was dropped. Currently this code path can still be
+        // invoked in the non-GPL stack, which works on AMD SEV-SNP processors
+        // as long as KVM is the hypervisor, so the assertion is skipped until
+        // the non-GPL case is officially deprecated.
+#endif
+
+   return prot;
+}
+
 // Commit 1dff8083a024650c75a9c961c38082473ceae8cf (v4.7).
 //
 // Archs with CONFIG_MMU should have their own page.h, and can't include
