@@ -99,7 +99,7 @@ static void nv_init_dynamic_power_management
                    NV_PCI_DOMAIN_NUMBER(pci_dev),
                    NV_PCI_BUS_NUMBER(pci_dev),
                    NV_PCI_SLOT_NUMBER(pci_dev));
-    if (ret > 0 || ret < sizeof(filename))
+    if (ret > 0 && ret < sizeof(filename))
     {
         struct file *file = filp_open(filename, O_RDONLY, 0);
         if (!IS_ERR(file))
@@ -144,80 +144,31 @@ static void nv_init_dynamic_power_management
         }
     }
 
-    if (pci_dev->bus && pci_dev->bus->self)
+    if (nv_get_hypervisor_type() != OS_HYPERVISOR_UNKNOWN)
     {
-        pr3_acpi_method_present = nv_acpi_upstreamport_power_resource_method_present(pci_dev->bus->self);
+        pr3_acpi_method_present = nv_acpi_power_resource_method_present(pci_dev);
+    }
+    else if (pci_dev->bus && pci_dev->bus->self)
+    {
+        pr3_acpi_method_present = nv_acpi_power_resource_method_present(pci_dev->bus->self);
     }
 
     rm_init_dynamic_power_management(sp, nv, pr3_acpi_method_present);
 }
 
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
+static int nv_resize_pcie_bars(struct pci_dev *pci_dev) {
+    nv_printf(NV_DBG_INFO, "NVRM: Resizable BAR is not supported on this driver version.\n");
+    return 0;
+}
+
+static void
+nv_init_coherent_link_info
+(
+    nv_state_t *nv
+)
+{
+    return;
+}
 
 /* find nvidia devices and set initial state */
 static int
@@ -235,6 +186,8 @@ nv_pci_probe
     NvBool prev_nv_ats_supported = nv_ats_supported;
     NV_STATUS status;
     NvBool last_bar_64bit = NV_FALSE;
+    NvU8 regs_bar_index = nv_bar_index_to_os_bar_index(pci_dev,
+                                                       NV_GPU_BAR_INDEX_REGS);
 
     nv_printf(NV_DBG_SETUP, "NVRM: probing 0x%x 0x%x, class 0x%x\n",
         pci_dev->vendor, pci_dev->device, pci_dev->class);
@@ -244,7 +197,6 @@ nv_pci_probe
         return -1;
     }
 
-
 #ifdef NV_PCI_SRIOV_SUPPORT
     if (pci_dev->is_virtfn)
     {
@@ -290,7 +242,6 @@ nv_pci_probe
     }
 #endif /* NV_PCI_SRIOV_SUPPORT */
 
-
     if (!rm_is_supported_pci_device(
                 (pci_dev->class >> 16) & 0xFF,
                 (pci_dev->class >> 8) & 0xFF,
@@ -412,31 +363,43 @@ next_bar:
             (NvU64)NV_PCI_RESOURCE_START(pci_dev, i),
             NV_PCI_DOMAIN_NUMBER(pci_dev), NV_PCI_BUS_NUMBER(pci_dev),
             NV_PCI_SLOT_NUMBER(pci_dev), PCI_FUNC(pci_dev->devfn));
+
+        // With GH180 C2C, VF BAR1/2 are disabled and therefore expected to be 0.
+        if (j != NV_GPU_BAR_INDEX_REGS)
+        {
+            nv_printf(NV_DBG_INFO, "NVRM: ignore invalid BAR failure for BAR%d\n", j);
+            continue;
+        }
+
         goto failed;
     }
 
-    if (!request_mem_region(NV_PCI_RESOURCE_START(pci_dev, NV_GPU_BAR_INDEX_REGS),
-                            NV_PCI_RESOURCE_SIZE(pci_dev, NV_GPU_BAR_INDEX_REGS),
+    if (!request_mem_region(NV_PCI_RESOURCE_START(pci_dev, regs_bar_index),
+                            NV_PCI_RESOURCE_SIZE(pci_dev, regs_bar_index),
                             nv_device_name))
     {
         nv_printf(NV_DBG_ERRORS,
             "NVRM: request_mem_region failed for %dM @ 0x%llx. This can\n"
             "NVRM: occur when a driver such as rivatv is loaded and claims\n"
             "NVRM: ownership of the device's registers.\n",
-            (NV_PCI_RESOURCE_SIZE(pci_dev, NV_GPU_BAR_INDEX_REGS) >> 20),
-            (NvU64)NV_PCI_RESOURCE_START(pci_dev, NV_GPU_BAR_INDEX_REGS));
+            (NV_PCI_RESOURCE_SIZE(pci_dev, regs_bar_index) >> 20),
+            (NvU64)NV_PCI_RESOURCE_START(pci_dev, regs_bar_index));
         goto failed;
     }
 
-    NV_KMALLOC(nvl, sizeof(nv_linux_state_t));
+    if (nv_resize_pcie_bars(pci_dev)) {
+        nv_printf(NV_DBG_ERRORS,
+            "NVRM: Fatal Error while attempting to resize PCIe BARs.\n");
+        goto failed;
+    }
+
+    NV_KZALLOC(nvl, sizeof(nv_linux_state_t));
     if (nvl == NULL)
     {
         nv_printf(NV_DBG_ERRORS, "NVRM: failed to allocate memory\n");
         goto err_not_supported;
     }
 
-    os_mem_set(nvl, 0, sizeof(nv_linux_state_t));
-
     nv  = NV_STATE_PTR(nvl);
 
     pci_set_drvdata(pci_dev, (void *)nvl);
@@ -478,8 +441,7 @@ next_bar:
         {
             nv->bars[j].offset = NVRM_PCICFG_BAR_OFFSET(i);
             nv->bars[j].cpu_address = NV_PCI_RESOURCE_START(pci_dev, i);
-            nv->bars[j].strapped_size = NV_PCI_RESOURCE_SIZE(pci_dev, i);
-            nv->bars[j].size = nv->bars[j].strapped_size;
+            nv->bars[j].size = NV_PCI_RESOURCE_SIZE(pci_dev, i);
             j++;
         }
     }
@@ -493,20 +455,16 @@ next_bar:
 
     nv_init_ibmnpu_info(nv);
 
-
-
-
+    nv_init_coherent_link_info(nv);
 
 #if defined(NVCPU_PPC64LE)
     // Use HW NUMA support as a proxy for ATS support. This is true in the only
     // PPC64LE platform where ATS is currently supported (IBM P9).
     nv_ats_supported &= nv_platform_supports_numa(nvl);
 #else
-
-
-
-
-
+#if defined(NV_PCI_DEV_HAS_ATS_ENABLED)
+    nv_ats_supported &= pci_dev->ats_enabled;
+#endif
 #endif
     if (nv_ats_supported)
     {
@@ -573,38 +531,51 @@ next_bar:
     if (nvidia_frontend_add_device((void *)&nv_fops, nvl) != 0)
         goto err_remove_device;
 
-#if defined(NV_PM_VT_SWITCH_REQUIRED_PRESENT)
     pm_vt_switch_required(nvl->dev, NV_TRUE);
-#endif
+
+    nv_init_dynamic_power_management(sp, pci_dev);
 
     nv_procfs_add_gpu(nvl);
 
     /* Parse and set any per-GPU registry keys specified. */
     nv_parse_per_device_option_string(sp);
 
+    rm_set_rm_firmware_requested(sp, nv);
+
 #if defined(NV_VGPU_KVM_BUILD)
     if (nvidia_vgpu_vfio_probe(nvl->pci_dev) != NV_OK)
     {
         NV_DEV_PRINTF(NV_DBG_ERRORS, nv, "Failed to register device to vGPU VFIO module");
         nvidia_frontend_remove_device((void *)&nv_fops, nvl);
-        goto err_remove_device;
+        goto err_vgpu_kvm;
     }
 #endif
 
     nv_check_and_exclude_gpu(sp, nv);
 
-    rm_set_rm_firmware_requested(sp, nv);
-
-    nv_init_dynamic_power_management(sp, pci_dev);
-
-#if defined(DPM_FLAG_NEVER_SKIP)
+#if defined(DPM_FLAG_NO_DIRECT_COMPLETE)
+    dev_pm_set_driver_flags(nvl->dev, DPM_FLAG_NO_DIRECT_COMPLETE);
+#elif defined(DPM_FLAG_NEVER_SKIP)
     dev_pm_set_driver_flags(nvl->dev, DPM_FLAG_NEVER_SKIP);
 #endif
 
+    /*
+     * Dynamic power management should be enabled as the last step.
+     * Kernel runtime power management framework can put the device
+     * into the suspended state. Hardware register access should not be done
+     * after enabling dynamic power management.
+     */
+    rm_enable_dynamic_power_management(sp, nv);
     nv_kmem_cache_free_stack(sp);
 
     return 0;
 
+#if defined(NV_VGPU_KVM_BUILD)
+err_vgpu_kvm:
+#endif
+    nv_procfs_remove_gpu(nvl);
+    rm_cleanup_dynamic_power_management(sp, nv);
+    pm_vt_switch_unregister(nvl->dev);
 err_remove_device:
     LOCK_NV_LINUX_DEVICES();
     nv_linux_remove_device_locked(nvl);
@@ -619,8 +590,8 @@ err_not_supported:
     {
         NV_KFREE(nvl, sizeof(nv_linux_state_t));
     }
-    release_mem_region(NV_PCI_RESOURCE_START(pci_dev, NV_GPU_BAR_INDEX_REGS),
-                       NV_PCI_RESOURCE_SIZE(pci_dev, NV_GPU_BAR_INDEX_REGS));
+    release_mem_region(NV_PCI_RESOURCE_START(pci_dev, regs_bar_index),
+                       NV_PCI_RESOURCE_SIZE(pci_dev, regs_bar_index));
     NV_PCI_DISABLE_DEVICE(pci_dev);
     pci_set_drvdata(pci_dev, NULL);
 failed:
@@ -634,12 +605,13 @@ nv_pci_remove(struct pci_dev *pci_dev)
     nv_linux_state_t *nvl = NULL;
     nv_state_t *nv;
     nvidia_stack_t *sp = NULL;
+    NvU8 regs_bar_index = nv_bar_index_to_os_bar_index(pci_dev,
+                                                       NV_GPU_BAR_INDEX_REGS);
 
     nv_printf(NV_DBG_SETUP, "NVRM: removing GPU %04x:%02x:%02x.%x\n",
               NV_PCI_DOMAIN_NUMBER(pci_dev), NV_PCI_BUS_NUMBER(pci_dev),
               NV_PCI_SLOT_NUMBER(pci_dev), PCI_FUNC(pci_dev->devfn));
 
-
 #ifdef NV_PCI_SRIOV_SUPPORT
     if (pci_dev->is_virtfn)
     {
@@ -651,7 +623,6 @@ nv_pci_remove(struct pci_dev *pci_dev)
     }
 #endif /* NV_PCI_SRIOV_SUPPORT */
 
-
     if (nv_kmem_cache_alloc_stack(&sp) != 0)
     {
         return;
@@ -675,8 +646,9 @@ nv_pci_remove(struct pci_dev *pci_dev)
     if ((NV_ATOMIC_READ(nvl->usage_count) != 0) && !(nv->is_external_gpu))
     {
         nv_printf(NV_DBG_ERRORS,
-                  "NVRM: Attempting to remove minor device %u with non-zero usage count!\n",
-                  nvl->minor_num);
+                  "NVRM: Attempting to remove device %04x:%02x:%02x.%x with non-zero usage count!\n",
+                  NV_PCI_DOMAIN_NUMBER(pci_dev), NV_PCI_BUS_NUMBER(pci_dev),
+                  NV_PCI_SLOT_NUMBER(pci_dev), PCI_FUNC(pci_dev->devfn));
 
         /*
          * We can't return from this function without corrupting state, so we wait for
@@ -701,8 +673,9 @@ nv_pci_remove(struct pci_dev *pci_dev)
             {
                 /* The device was not found, which should not happen */
                 nv_printf(NV_DBG_ERRORS,
-                          "NVRM: Failed removal of minor device %u!\n",
-                          nvl->minor_num);
+                          "NVRM: Failed removal of device %04x:%02x:%02x.%x!\n",
+                          NV_PCI_DOMAIN_NUMBER(pci_dev), NV_PCI_BUS_NUMBER(pci_dev),
+                          NV_PCI_SLOT_NUMBER(pci_dev), PCI_FUNC(pci_dev->devfn));  
                 WARN_ON(1);
                 goto done;
             }
@@ -711,12 +684,11 @@ nv_pci_remove(struct pci_dev *pci_dev)
         }
 
         nv_printf(NV_DBG_ERRORS,
-                  "NVRM: Continuing with GPU removal for minor device %u\n",
-                  nvl->minor_num);
+                  "NVRM: Continuing with GPU removal for device %04x:%02x:%02x.%x\n",
+                  NV_PCI_DOMAIN_NUMBER(pci_dev), NV_PCI_BUS_NUMBER(pci_dev),
+                  NV_PCI_SLOT_NUMBER(pci_dev), PCI_FUNC(pci_dev->devfn));
     }
 
-    rm_cleanup_dynamic_power_management(sp, nv);
-
     rm_check_for_gpu_surprise_removal(sp, nv);
 
     nv_linux_remove_device_locked(nvl);
@@ -724,13 +696,13 @@ nv_pci_remove(struct pci_dev *pci_dev)
     /* Remove proc entry for this GPU */
     nv_procfs_remove_gpu(nvl);
 
+    rm_cleanup_dynamic_power_management(sp, nv);
+
     nv->removed = NV_TRUE;
 
     UNLOCK_NV_LINUX_DEVICES();
 
-#if defined(NV_PM_VT_SWITCH_REQUIRED_PRESENT)
     pm_vt_switch_unregister(&pci_dev->dev);
-#endif
 
 #if defined(NV_VGPU_KVM_BUILD)
     /* Arg 2 == NV_TRUE means that the PCI device should be removed */
@@ -774,8 +746,8 @@ nv_pci_remove(struct pci_dev *pci_dev)
 
     rm_i2c_remove_adapters(sp, nv);
     rm_free_private_state(sp, nv);
-    release_mem_region(NV_PCI_RESOURCE_START(pci_dev, NV_GPU_BAR_INDEX_REGS),
-                       NV_PCI_RESOURCE_SIZE(pci_dev, NV_GPU_BAR_INDEX_REGS));
+    release_mem_region(NV_PCI_RESOURCE_START(pci_dev, regs_bar_index),
+                       NV_PCI_RESOURCE_SIZE(pci_dev, regs_bar_index));
 
     num_nv_devices--;
 
@@ -808,6 +780,11 @@ nv_pci_shutdown(struct pci_dev *pci_dev)
         return;
     }
 
+    if (nvl != NULL)
+    {
+        nvl->nv_state.is_shutdown = NV_TRUE;
+    }
+
     /* pci_clear_master is not defined for !CONFIG_PCI */
 #ifdef CONFIG_PCI
     pci_clear_master(pci_dev);
@@ -1057,6 +1034,10 @@ struct pci_driver nv_pci_driver = {
     .probe     = nv_pci_probe,
     .remove    = nv_pci_remove,
     .shutdown  = nv_pci_shutdown,
+#if defined(NV_USE_VFIO_PCI_CORE) && \
+  defined(NV_PCI_DRIVER_HAS_DRIVER_MANAGED_DMA)
+    .driver_managed_dma = NV_TRUE,
+#endif
 #if defined(CONFIG_PM)
     .driver.pm = &nv_pm_ops,
 #endif
