@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 1999-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 1999-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -41,8 +41,8 @@ extern nv_kthread_q_t nv_kthread_q;
 NvU32 os_page_size  = PAGE_SIZE;
 NvU64 os_page_mask  = NV_PAGE_MASK;
 NvU8  os_page_shift = PAGE_SHIFT;
-NvU32 os_sev_status = 0;
-NvBool os_sev_enabled = 0;
+NvBool os_cc_enabled = 0;
+NvBool os_cc_tdx_enabled = 0;
 
 NvBool os_dma_buf_enabled = NV_FALSE;
 
@@ -169,7 +169,7 @@ void* NV_API_CALL os_alloc_semaphore
         return NULL;
     }
 
-    NV_INIT_SEMA(os_sema, initialValue);
+    sema_init(os_sema, initialValue);
 
     return (void *)os_sema;
 }
@@ -229,6 +229,90 @@ NV_STATUS NV_API_CALL os_release_semaphore
     return NV_OK;
 }
 
+typedef struct rw_semaphore os_rwlock_t;
+
+void* NV_API_CALL os_alloc_rwlock(void)
+{
+    os_rwlock_t *os_rwlock = NULL;
+
+    NV_STATUS rmStatus = os_alloc_mem((void *)&os_rwlock, sizeof(os_rwlock_t));
+    if (rmStatus != NV_OK)
+    {
+        nv_printf(NV_DBG_ERRORS, "NVRM: failed to allocate rw_semaphore!\n");
+        return NULL;
+    }
+
+    init_rwsem(os_rwlock);
+
+    return os_rwlock;
+}
+
+void NV_API_CALL os_free_rwlock(void *pRwLock)
+{
+    os_rwlock_t *os_rwlock = (os_rwlock_t *)pRwLock;
+    os_free_mem(os_rwlock);
+}
+
+NV_STATUS NV_API_CALL os_acquire_rwlock_read(void *pRwLock)
+{
+    os_rwlock_t *os_rwlock = (os_rwlock_t *)pRwLock;
+
+    if (!NV_MAY_SLEEP())
+    {
+        return NV_ERR_INVALID_REQUEST;
+    }
+    down_read(os_rwlock);
+    return NV_OK;
+}
+
+NV_STATUS NV_API_CALL os_acquire_rwlock_write(void *pRwLock)
+{
+    os_rwlock_t *os_rwlock = (os_rwlock_t *)pRwLock;
+
+    if (!NV_MAY_SLEEP())
+    {
+        return NV_ERR_INVALID_REQUEST;
+    }
+    down_write(os_rwlock);
+    return NV_OK;
+}
+
+NV_STATUS NV_API_CALL os_cond_acquire_rwlock_read(void *pRwLock)
+{
+    os_rwlock_t *os_rwlock = (os_rwlock_t *)pRwLock;
+
+    if (down_read_trylock(os_rwlock))
+    {
+        return NV_ERR_TIMEOUT_RETRY;
+    }
+
+    return NV_OK;
+}
+
+NV_STATUS NV_API_CALL os_cond_acquire_rwlock_write(void *pRwLock)
+{
+    os_rwlock_t *os_rwlock = (os_rwlock_t *)pRwLock;
+
+    if (down_write_trylock(os_rwlock))
+    {
+        return NV_ERR_TIMEOUT_RETRY;
+    }
+
+    return NV_OK;
+}
+
+void NV_API_CALL os_release_rwlock_read(void *pRwLock)
+{
+    os_rwlock_t *os_rwlock = (os_rwlock_t *)pRwLock;
+    up_read(os_rwlock);
+}
+
+void NV_API_CALL os_release_rwlock_write(void *pRwLock)
+{
+    os_rwlock_t *os_rwlock = (os_rwlock_t *)pRwLock;
+    up_write(os_rwlock);
+}
+
 NvBool NV_API_CALL os_semaphore_may_sleep(void)
 {
     return NV_MAY_SLEEP();
@@ -469,6 +553,7 @@ NV_STATUS NV_API_CALL os_alloc_mem(
     NvU64 size
 )
 {
+    NvU64 original_size = size;
     unsigned long alloc_size;
 
     if (address == NULL)
@@ -477,6 +562,10 @@ NV_STATUS NV_API_CALL os_alloc_mem(
     *address = NULL;
     NV_MEM_TRACKING_PAD_SIZE(size);
 
+    // check for integer overflow on size
+    if (size < original_size)
+        return NV_ERR_INVALID_ARGUMENT;
+
     //
     // NV_KMALLOC, nv_vmalloc take an input of 4 bytes in x86. To avoid
     // truncation and wrong allocation, below check is required.
@@ -511,7 +600,7 @@ NV_STATUS NV_API_CALL os_alloc_mem(
 
 void NV_API_CALL os_free_mem(void *address)
 {
-    NvU32 size;
+    NvU64 size;
 
     NV_MEM_TRACKING_RETRIEVE_SIZE(address, size);
 
@@ -1077,14 +1166,14 @@ void NV_API_CALL os_dbg_breakpoint(void)
 #endif // CONFIG_X86_REMOTE_DEBUG || CONFIG_KGDB || CONFIG_XMON
 }
 
-NvU32 NV_API_CALL os_get_cpu_number()
+NvU32 NV_API_CALL os_get_cpu_number(void)
 {
     NvU32 cpu_id = get_cpu();
     put_cpu();
     return cpu_id;
 }
 
-NvU32 NV_API_CALL os_get_cpu_count()
+NvU32 NV_API_CALL os_get_cpu_count(void)
 {
     return NV_NUM_CPUS();
 }
@@ -1096,7 +1185,7 @@ NvBool NV_API_CALL os_pat_supported(void)
 
 NvBool NV_API_CALL os_is_efi_enabled(void)
 {
-    return NV_EFI_ENABLED();
+    return efi_enabled(EFI_BOOT);
 }
 
 void NV_API_CALL os_get_screen_info(
@@ -1131,11 +1220,27 @@ void NV_API_CALL os_get_screen_info(
                 *pFbHeight = registered_fb[i]->var.yres;
                 *pFbDepth = registered_fb[i]->var.bits_per_pixel;
                 *pFbPitch = registered_fb[i]->fix.line_length;
-                break;
+                return;
             }
         }
     }
-#elif NV_IS_EXPORT_SYMBOL_PRESENT_screen_info
+#endif
+
+    /*
+     * If the screen info is not found in the registered FBs then fallback
+     * to the screen_info structure.
+     *
+     * The SYSFB_SIMPLEFB option, if enabled, marks VGA/VBE/EFI framebuffers as
+     * generic framebuffers so the new generic system-framebuffer drivers can
+     * be used instead. DRM_SIMPLEDRM drives the generic system-framebuffers
+     * device created by SYSFB_SIMPLEFB.
+     *
+     * SYSFB_SIMPLEFB registers a dummy framebuffer which does not contain the
+     * information required by os_get_screen_info(), therefore you need to
+     * fall back onto the screen_info structure.
+     */
+
+#if NV_IS_EXPORT_SYMBOL_PRESENT_screen_info
     /*
      * If there is not a framebuffer console, return 0 size.
      *
@@ -1164,7 +1269,7 @@ void NV_API_CALL os_get_screen_info(
 #endif
 }
 
-void NV_API_CALL os_dump_stack()
+void NV_API_CALL os_dump_stack(void)
 {
     dump_stack();
 }
@@ -1313,8 +1418,7 @@ NV_STATUS NV_API_CALL os_get_euid(NvU32 *pSecToken)
     return NV_OK;
 }
 
-// These functions are needed only on x86_64 platforms.
-#if defined(NVCPU_X86_64)
+#if defined(NVCPU_X86_64) || defined(NVCPU_AARCH64)
 
 static NvBool os_verify_checksum(const NvU8 *pMappedAddr, NvU32 length)
 {
@@ -1352,6 +1456,9 @@ static NvBool os_verify_checksum(const NvU8 *pMappedAddr, NvU32 length)
 
 static NV_STATUS os_get_smbios_header_legacy(NvU64 *pSmbsAddr)
 {
+#if !defined(NVCPU_X86_64)
+    return NV_ERR_NOT_SUPPORTED;
+#else
     NV_STATUS status = NV_ERR_OPERATING_SYSTEM;
     NvU8 *pMappedAddr = NULL;
     NvU8 *pIterAddr = NULL;
@@ -1386,6 +1493,7 @@ static NV_STATUS os_get_smbios_header_legacy(NvU64 *pSmbsAddr)
     os_unmap_kernel_space(pMappedAddr, SMBIOS_LEGACY_SIZE);
 
     return status;
+#endif
 }
 
 // This function is needed only if "efi" is enabled.
@@ -1462,13 +1570,13 @@ static NV_STATUS os_get_smbios_header_uefi(NvU64 *pSmbsAddr)
     return status;
 }
 
-#endif // defined(NVCPU_X86_64)
+#endif // defined(NVCPU_X86_64) || defined(NVCPU_AARCH64)
 
 // The function locates the SMBIOS entry point.
 NV_STATUS NV_API_CALL os_get_smbios_header(NvU64 *pSmbsAddr)
 {
 
-#if !defined(NVCPU_X86_64)
+#if !defined(NVCPU_X86_64) && !defined(NVCPU_AARCH64)
     return NV_ERR_NOT_SUPPORTED;
 #else
     NV_STATUS status = NV_OK;
@@ -1681,11 +1789,6 @@ NV_STATUS NV_API_CALL os_numa_memblock_size
     return NV_OK;
 }
 
-NV_STATUS NV_API_CALL os_call_nv_vmbus(NvU32 vmbus_cmd, void *input)
-{
-    return NV_ERR_NOT_SUPPORTED;
-}
-
 NV_STATUS NV_API_CALL os_open_temporary_file
 (
     void **ppFile
@@ -1756,7 +1859,6 @@ NV_STATUS NV_API_CALL os_write_file
     NvU64 offset
 )
 {
-#if defined(NV_KERNEL_WRITE_PRESENT)
     loff_t f_pos = offset;
     ssize_t num_written;
     int num_retries = NV_MAX_NUM_FILE_IO_RETRIES;
@@ -1787,9 +1889,6 @@ retry:
     }
 
     return NV_OK;
-#else
-    return NV_ERR_NOT_SUPPORTED;
-#endif
 }
 
 NV_STATUS NV_API_CALL os_read_file
@@ -1898,13 +1997,22 @@ NvBool NV_API_CALL os_is_nvswitch_present(void)
     return !!pci_dev_present(nvswitch_pci_table);
 }
 
-void NV_API_CALL os_get_random_bytes
+/*
+ * This function may sleep (interruptible).
+ */
+NV_STATUS NV_API_CALL os_get_random_bytes
 (
     NvU8 *bytes,
     NvU16 numBytes
 )
 {
+#if defined NV_WAIT_FOR_RANDOM_BYTES_PRESENT
+    if (wait_for_random_bytes() < 0)
+        return NV_ERR_NOT_READY;
+#endif
+
     get_random_bytes(bytes, numBytes);
+    return NV_OK;
 }
 
 NV_STATUS NV_API_CALL os_alloc_wait_queue
@@ -2006,3 +2114,80 @@ void NV_API_CALL os_nv_cap_close_fd
     nv_cap_close_fd(fd);
 }
 
+typedef struct os_numa_gpu_mem_hotplug_notifier_s
+{
+    NvU64 start_pa;
+    NvU64 size;
+    nv_pci_info_t pci_info;
+    struct notifier_block memory_notifier;
+} os_numa_gpu_mem_hotplug_notifier_t;
+
+static int os_numa_verify_gpu_memory_zone(struct notifier_block *nb,
+                                          unsigned long action, void *data)
+{
+    os_numa_gpu_mem_hotplug_notifier_t *notifier = container_of(nb,
+        os_numa_gpu_mem_hotplug_notifier_t,
+        memory_notifier);
+    struct memory_notify *mhp = data;
+    NvU64 start_pa = PFN_PHYS(mhp->start_pfn);
+    NvU64 size = PFN_PHYS(mhp->nr_pages);
+
+    if (action == MEM_GOING_ONLINE)
+    {
+        // Check if onlining memory falls in the GPU memory range
+        if ((start_pa >= notifier->start_pa) &&
+            (start_pa + size) <= (notifier->start_pa + notifier->size))
+        {
+            /*
+             * Verify GPU memory NUMA node has memory only in ZONE_MOVABLE before
+             * onlining the memory so that incorrect auto online setting doesn't
+             * cause the memory onlined in a zone where kernel allocations
+             * could happen, resulting in GPU memory hot unpluggable and requiring
+             * system reboot.
+             */
+            if (page_zonenum((pfn_to_page(mhp->start_pfn))) != ZONE_MOVABLE)
+            {
+                nv_printf(NV_DBG_ERRORS, "NVRM: Failing GPU memory onlining as the onlining zone "
+                          "is not movable. pa: 0x%llx size: 0x%llx\n"
+                          "NVRM: The NVIDIA GPU %04x:%02x:%02x.%x installed in the system\n"
+                          "NVRM: requires auto onlining mode online_movable enabled in\n"
+                          "NVRM: /sys/devices/system/memory/auto_online_blocks\n",
+                          start_pa, size, notifier->pci_info.domain, notifier->pci_info.bus,
+                          notifier->pci_info.slot, notifier->pci_info.function);
+                return NOTIFY_BAD;
+            }
+        }
+    }
+    return NOTIFY_OK;
+}
+
+NV_STATUS NV_API_CALL os_numa_add_gpu_memory
+(
+    void *handle,
+    NvU64 offset,
+    NvU64 size,
+    NvU32 *nodeId
+)
+{
+    return NV_ERR_NOT_SUPPORTED;
+}
+
+NV_STATUS NV_API_CALL os_numa_remove_gpu_memory
+(
+    void *handle,
+    NvU64 offset,
+    NvU64 size,
+    NvU32 nodeId
+)
+{
+    return NV_ERR_NOT_SUPPORTED;
+}
+
+NV_STATUS NV_API_CALL os_offline_page_at_address
+(
+    NvU64 address
+)
+{
+    return NV_ERR_NOT_SUPPORTED;
+}
+
